import warnings

import matplotlib.pyplot as plt
import numpy as np
import pytest

from astropy.io import fits

import sunkit_image.data.test as data
from sunkit_image.tests.helpers import figure_test
from sunkit_image.trace import (
    bandpass_filter,
    curvature_radius,
    erase_loop_in_image,
    initial_direction_finding,
    loop_add,
    occult2,
    smooth,
)


@pytest.fixture
@pytest.mark.remote_data
def image_remote():
    with warnings.catch_warnings():
        warnings.simplefilter("ignore", category=fits.verify.VerifyWarning)
        im = fits.getdata("http://data.sunpy.org/sunkit-image/trace_1998-05-19T22:21:43.000_171_1024.fits")
        return im


@pytest.fixture
def filepath_IDL():

    filepath = data.get_test_filepath("IDL.txt")
    return filepath


@pytest.mark.remote_data
def test_occult2_remote(image_remote, filepath_IDL):

    # Testing on the same input files as in the IDL tutorial
    loops = occult2(image_remote, nsm1=3, rmin=30, lmin=25, nstruc=1000, ngap=0, qthresh1=0.0, qthresh2=3.0)

    # Taking all the x and y coordinates in seperate lists
    x = []
    y = []
    for loop in loops:
        for points in loop:
            x.append(points[0])
            y.append(points[1])

    # Creating a numpy array of all the loop points for ease of comparison
    X = np.array(x)
    Y = np.array(y)
    coords_py = np.c_[X, Y]

    # Now we will test on the IDL output data

    # Reading the IDL file
    expect = np.loadtxt(filepath_IDL)

    # Validating the number of loops
    assert np.allclose(expect[-1, 0] + 1, len(loops))

    # We know that the python code detects one point extra than the IDL code.
    # So to test it we will remove that point.
    coords_py = np.delete(coords_py, (1745), axis=0)

    # Taking all the coords from the IDL form
    coords_idl = expect[:, 1:3]

    # Checking all the coordinates must be close to each other
    assert np.allclose(coords_py, coords_idl, atol=1e-0, rtol=1e-10)

    # We devise one more test where we will find the distance between the Python and IDL points
    # For the algorithm to work correctly this distance should be very small.
    diff = coords_idl - coords_py
    square_diff = diff ** 2
    sum_diff = np.sum(square_diff, axis=1)

    distance = np.sqrt(sum_diff)

    # The maximum distance between the IDL points and the Python points was found to be 0.11 pixels.
    assert all(distance < 0.11)


@figure_test
@pytest.mark.remote_data
def test_occult2_fig(image_remote):

    # A figure test for occult2, the plot is same as the one in the IDL tutorial
    loops = occult2(image_remote, nsm1=3, rmin=30, lmin=25, nstruc=1000, ngap=0, qthresh1=0.0, qthresh2=3.0)

    for loop in loops:

        # We collect all the ``x`` and ``y`` coordinates in seperate lists for plotting.
        x = []
        y = []
        for points in loop:
            x.append(points[0])
            y.append(points[1])

        plt.plot(x, y, "b")


@pytest.fixture
def test_image():

    # An image containing a loop in a straight line
    ima = np.zeros((3, 3), dtype=np.float32)
    ima[0, 1] = 5
    ima[1, 1] = 3
    ima[2, 1] = 0
    return ima


@pytest.fixture
def image_test():

    # An image containing a loop in a straight line
    ima = np.zeros((15, 15), dtype=np.float32)
    ima[:, 7] = 1
    ima[3:12, 7] = [4, 3, 6, 12, 4, 3, 4, 2, 1]

    return ima


def test_occult2(test_image, image_test):

    # Set of checks which does not require remote data

    # The first test were valid loops are detected
    loops = occult2(image_test, nsm1=1, rmin=30, lmin=0, nstruc=1000, ngap=0, qthresh1=0.0, qthresh2=3.0)

    for loop in loops:

        # We collect all the ``x`` and ``y`` coordinates in seperate lists
        x = []
        y = []
        for points in loop:
            x.append(points[0])
            y.append(points[1])

    # From the input image it is clear that all x coordinate is 7.
    assert np.allclose(np.round(x), np.ones(8) * 7)
    # All the y coords are [11, 10, ..., 4]
    assert np.allclose(np.round(y), np.arange(11, 3, -1))

    # This check will return an empty list as no loop is detected
    loops = occult2(image_test, nsm1=1, rmin=30, lmin=25, nstruc=1000, ngap=0, qthresh1=0.0, qthresh2=3.0)
    assert not loops

    # This check is used to verify whether the RuntimeError is triggered
    with pytest.raises(RuntimeError) as record:
        _ = occult2(test_image, nsm1=3, rmin=30, lmin=25, nstruc=1000, ngap=0, qthresh1=0.0, qthresh2=3.0)

    assert (
        str(record.value)
        == "The filter size is very large compared to the size of the image."
        + " The entire image zeros out while smoothing the image edges after filtering."
    )


@pytest.fixture
def test_map():
    map_test = [[1.0, 1.0, 1.0, 1.0], [1.0, 5.0, 5.0, 1.0], [1.0, 5.0, 5.0, 1.0], [1.0, 1.0, 1.0, 1.0]]

    return np.array(map_test)


@pytest.fixture
def image():
    return np.ones((4, 4), dtype=np.float32)


def test_bandpass_filter(image, test_map):

    expect = np.zeros((4, 4))
    result = bandpass_filter(image)

    assert np.allclose(expect, result)

    expect = np.array(
        [
            [0.0, 0.0, 0.0, 0.0],
            [0.0, 2.22222222, 2.22222222, 0.0],
            [0.0, 2.22222222, 2.22222222, 0.0],
            [0.0, 0.0, 0.0, 0.0],
        ]
    )

    result = bandpass_filter(test_map)

    assert np.allclose(expect, result)

    with pytest.raises(ValueError) as record:
        _ = bandpass_filter(image, 5, 1)

    assert str(record.value) == "nsm1 should be less than nsm2"


def test_smooth(image, test_map):

    filtered = smooth(image, 1)
    assert np.allclose(filtered, image)

    filtered = smooth(image, 4)
    assert np.allclose(filtered, image)

    filtered = smooth(test_map, 1)
    assert np.allclose(filtered, test_map)

    filtered = smooth(test_map, 3)
    expect = np.array(
        [
            [1.0, 1.0, 1.0, 1.0],
            [1.0, 2.77777777, 2.77777777, 1.0],
            [1.0, 2.77777777, 2.77777777, 1.0],
            [1.0, 1.0, 1.0, 1.0],
        ]
    )

    assert np.allclose(filtered, expect)


def test_erase_loop_in_image(image, test_map):

    # The starting point of a dummy loop
    istart = 0
    jstart = 1
    width = 1

    # The coordinates of the dummy loop
    xloop = [1, 2, 3]
    yloop = [1, 1, 1]

    result = erase_loop_in_image(image, istart, jstart, width, xloop, yloop)

    expect = np.array(
        [[0.0, 0.0, 0.0, 1.0], [0.0, 0.0, 0.0, 1.0], [0.0, 0.0, 0.0, 1.0], [0.0, 0.0, 0.0, 1.0]]
    )

    assert np.allclose(expect, result)

    result = erase_loop_in_image(test_map, istart, jstart, width, xloop, yloop)

    expect = np.array(
        [[0.0, 0.0, 0.0, 1.0], [0.0, 0.0, 0.0, 1.0], [0.0, 0.0, 0.0, 1.0], [0.0, 0.0, 0.0, 1.0]]
    )

    assert np.allclose(expect, result)


@pytest.fixture
def test_image():

    # An image containing a loop in a straight line
    ima = np.zeros((3, 3), dtype=np.float32)
    ima[0, 1] = 5
    ima[1, 1] = 3
    ima[2, 1] = 0
    return ima


def test_initial_direction_finding(test_image):

    # The starting point of the loop i.e. the maximumflux position
    xstart = 0
    ystart = 1
    nlen = 30

    # The angle returned is with respect to the ``x`` axis.
    al = initial_direction_finding(test_image, xstart, ystart, nlen)

    # The angle returned is zero because the image has loop in the ``y`` direction but the function
    # assumes the image is transposed so it takes the straight line in the ``x`` direction.
    assert np.allclose(al, 0.0)


def test_curvature_radius(test_image):

    xl = np.zeros((3), dtype=np.float32)
    yl = np.zeros((3), dtype=np.float32)
    zl = np.zeros((3), dtype=np.float32)
    al = np.zeros((3), dtype=np.float32)
    ir = np.zeros((3), dtype=np.float32)

    xl[0] = 0
    yl[0] = 1
    zl[0] = 5
    al[0] = 0.0

    # Using the similar settings in as in the IDL tutorial.
    # This is forward tracing where the first point is after the starting point is being traced.
    xl, yl, zl, al = curvature_radius(test_image, 30, xl, yl, zl, al, ir, 0, 30, 0)

    assert np.allclose(np.ceil(xl[1]), 1)
    assert np.allclose(np.ceil(yl[1]), 1)
    assert np.allclose(zl[1], 3)

    # This is forward tracing where the second point is after the starting point is being traced.
    xl, yl, zl, al = curvature_radius(test_image, 30, xl, yl, zl, al, ir, 1, 30, 0)

    assert np.allclose(np.ceil(xl[2]), 2)
    assert np.allclose(np.ceil(yl[2]), 1)
    assert np.allclose(zl[2], 0)


@pytest.fixture
def parameters_add_loop():

    # Here we are creating dummy coordinates and flux for a loop
    xloop = np.ones(8, dtype=np.float32) * 7
    yloop = np.arange(11, 3, -1, dtype=np.float32)
    zloop = np.array([1, 2, 4, 3, 4, 12, 6, 3], dtype=np.float32)

    iloop = 0
    np1 = len(xloop)

    # Calculate the length of each point
    lengths = np.zeros((np1), dtype=np.float32)

    for ip in range(1, np1):
        lengths[ip] = lengths[ip - 1] + np.sqrt(
            (xloop[ip] - xloop[ip - 1]) ** 2 + (yloop[ip] - yloop[ip - 1]) ** 2
        )

    # The empty structures in which the first loop is stored
    loops = []

    return (lengths, xloop, yloop, zloop, iloop, loops)


def test_add_loop(parameters_add_loop):

    # We call the add_loop function and the values should be placed in the structures
    loops, iloop = loop_add(*parameters_add_loop)

    expect_loops = [[[7.0, 11.0], [7.0, 10.0], [7.0, 9.0], [7.0, 8.0], [7.0, 7.0], [7.0, 6.0], [7.0, 5.0]]]

    assert np.allclose(loops, expect_loops)
    assert np.allclose(iloop, 1)


def test_parameters_add_loop(parameters_add_loop):

    lengths, xloop, yloop, zloop, iloop, loops = parameters_add_loop

    assert np.allclose(lengths, np.arange(0, 8))
    assert np.allclose(xloop, np.ones(8) * 7)
    assert np.allclose(yloop, np.arange(11, 3, -1))
    assert np.allclose(zloop, np.array([1, 2, 4, 3, 4, 12, 6, 3]))
    assert np.allclose(iloop, 0)
    assert not loops



import numpy as np
import pytest
from skimage import data

import sunkit_image.utils.noise as nf


@pytest.fixture
def img():
    image = data.camera()
    return image


def test_conv2d_matrix_size(img):

    tt = nf.conv2d_matrix(img, 11.0, 7.0)
    assert tt.shape == ((11 - img.shape[0] + 1) * (7 - img.shape[1] + 1), 11.0 * 7.0)


def test_noiselevel(img):

    noise_levels = np.array([5.0, 10.0, 20.0, 42.0])
    n_levels = np.zeros_like(noise_levels)
    n_patches = np.zeros_like(noise_levels)

    for n in range(noise_levels.size):
        noise = img + np.random.standard_normal(img.shape) * noise_levels[n]
        output = nf.noise_estimation(noise, patchsize=11, iterations=5)
        n_levels[n] = output["nlevel"]
        n_patches[n] = output["num"]

    assert np.abs(1 - n_levels.all() / noise_levels.all()) < 0.1
    assert all(n_patches > 10000.0)


def test_weak_texture_mask(img):

    noise_levels = 5
    noise = img + np.random.standard_normal(img.shape) * noise_levels
    output = nf.noise_estimation(noise, patchsize=11, iterations=5)

    assert np.sum(output["mask"]) / output["mask"].size < 1.0



#!/usr/bin/env python
import os
from itertools import chain

from extension_helpers import get_extensions
from setuptools import setup
from setuptools.config import read_configuration

################################################################################
# Programmatically generate some extras combos.
################################################################################
extras = read_configuration("setup.cfg")["options"]["extras_require"]

# Dev is everything
extras["dev"] = list(chain(*extras.values()))

# All is everything but tests and docs
exclude_keys = ("tests", "docs", "dev")
ex_extras = dict(filter(lambda i: i[0] not in exclude_keys, extras.items()))
# Concatenate all the values together for 'all'
extras["all"] = list(chain.from_iterable(ex_extras.values()))
setup(
    extras_require=extras,
    use_scm_version={"write_to": os.path.join("sunkit_image", "_version.py")},
    ext_modules=get_extensions(),
)



"""
==========================================
Computing Cross-Correlations and Time Lags
==========================================

This example shows how to compute cross-correlations
between light curves and map the resulting time lags,
those temporal offsets which maximize the cross-correlation
between the two signals, back to an image pixel.
This method was developed for studying temporal evolution of AIA intensities
by `Viall and Klimchuk (2012) <https://doi.org/10.1088/0004-637X/753/1/35>`__.
The specific implementation in this package is described in detail
in Appendix C of `Barnes et al. (2019) <https://doi.org/10.3847/1538-4357/ab290c>`__.
"""
# sphinx_gallery_thumbnail_number = 4
import dask.array
import matplotlib.pyplot as plt
import numpy as np
from mpl_toolkits.axes_grid1.axes_divider import make_axes_locatable

import astropy.units as u

from sunkit_image.time_lag import cross_correlation, get_lags, max_cross_correlation, time_lag

###################################################################
# Consider two timeseries whose peaks are separated in time by some
# interval. We will create a toy model with two Gaussian pulses. In
# practice, this method is often applied to many AIA light curves
# and is described in detail in
# `Viall and Klimchuk (2012) <https://doi.org/10.1088/0004-637X/753/1/35>`__.


def gaussian_pulse(x, x0, sigma):
    return np.exp(-((x - x0) ** 2) / (2 * sigma ** 2))


time = np.linspace(0, 1, 500) * u.s
s_a = gaussian_pulse(time, 0.4 * u.s, 0.02 * u.s)
s_b = gaussian_pulse(time, 0.6 * u.s, 0.02 * u.s)
plt.plot(time, s_a, label="A")
plt.plot(time, s_b, label="B")
plt.xlabel("Time [s]")
plt.ylabel("Signal")
plt.legend()

###################################################################
# The problem we are concerned with is how much do we need shift
# signal A, either forward or backward in time, in order for it to
# best line up with signal B. In other words, what is the time lag
# between A and B. In the context of analyzing light curves from
# AIA, this gives us a proxy for the cooling time between two
# narrowband channels and thus two temperatures. To find this,
# we can compute the cross-correlation between the two signals
# and find which "lag" yields the highest correlation.

lags = get_lags(time)
cc = cross_correlation(s_a, s_b, lags)
plt.plot(lags, cc)
plt.xlabel("Lag [s]")
plt.ylabel("Cross-correlation, AB")
plt.show()

###################################################################
# Additionally, we can also easily calculate the maximum value of the
# cross-correlation and the associate lag, or the time lag.

tl = time_lag(s_a, s_b, time)
max_cc = max_cross_correlation(s_a, s_b, time)
plt.plot(lags, cc)
plt.plot(tl, max_cc, marker="o", ls="", markersize=4)
plt.xlabel("Lag [s]")
plt.ylabel("Cross-correlation, AB")
plt.show()

###################################################################
# As expected from the first intensity plot, we find that the lag
# which maximizes the cross-correlation is approximately the separation
# between the mean values of the Gaussian pulses.

print("Time lag, A -> B = ", tl)

###################################################################
# Note that a positive time lag indicates that signal A has to be
# shifted forward in time to match signal B. By reversing the order
# of the inputs, we also reverse the sign of the time lag.

print("Time lag, B -> A =", time_lag(s_b, s_a, time))

###################################################################
# The real power in the time lag approach is it's ability to reveal
# large scale patterns of cooling in images of the Sun, particularly
# in active regions. All of these functions can also be applied to
# intensity data cubes to create a "time lag map".
#
# As an example, we'll create a fake data cube by repeating Gaussian
# pulses with varying means and then add some noise to them

means_a = np.tile(np.random.rand(10, 10), time.shape + (1, 1)) * u.s
means_b = np.tile(np.random.rand(10, 10), time.shape + (1, 1)) * u.s
noise = 0.2 * (-0.5 + np.random.rand(*means_a.shape))
s_a = gaussian_pulse(np.tile(time, means_a.shape[1:] + (1,)).T, means_a, 0.02 * u.s) + noise
s_b = gaussian_pulse(np.tile(time, means_b.shape[1:] + (1,)).T, means_b, 0.02 * u.s) + noise

###################################################################
# We can now compute a map of the time lag and maximum cross correlation.

max_cc_map = max_cross_correlation(s_a, s_b, time)
tl_map = time_lag(s_a, s_b, time)

fig = plt.figure(figsize=(10, 5))

ax = fig.add_subplot(121)
im = ax.imshow(tl_map, cmap="RdBu", vmin=-1, vmax=1)
cax = make_axes_locatable(ax).append_axes("right", size="5%", pad="1%")
cb = fig.colorbar(im, cax=cax)
cb.set_label(r"$\tau_{AB}$ [s]")
ax = fig.add_subplot(122)
im = ax.imshow(max_cc_map, vmin=0, vmax=1, cmap="magma")
cax = make_axes_locatable(ax).append_axes("right", size="5%", pad="1%")
cb = fig.colorbar(im, cax=cax)
cb.set_label(r"Max cross-correlation")
plt.tight_layout()
plt.show()

###################################################################
# In practice, these data cubes are often very large, sometimes many
# GB, such that doing operations like these on them can be prohibitively
# expensive. All of these operations can be parallelized and distributed
# easily by passing in the intensity cubes as Dask arrays.

s_a = dask.array.from_array(s_a, chunks=s_a.shape[:1] + (5, 5))
s_b = dask.array.from_array(s_b, chunks=s_b.shape[:1] + (5, 5))
tl_map = time_lag(s_a, s_b, time)
print(tl_map)

###################################################################
# Rather than being computed "eagerly", :func:`~sunkit_image.time_lag.time_lag` returns
# a graph of the computation that can be handed off to a distributed
# scheduler to be run in parallel. This is extremely advantageous for
# large data cubes as these operations are likely to exceed the
# memory limits of most desktop machines and are easily accelerated through
# parallelism.
